@@ -6,6 +6,7 @@ from os.path import dirname, exists
 from subprocess import check_call
 from dataclasses import dataclass, field, fields
 from typing import Callable, List, Dict, Tuple, Union, Iterable, Literal, Optional, TYPE_CHECKING
+from time import sleep
 
 import numpy as np
 from obspy import Stream, Trace
@@ -13,12 +14,11 @@ from obspy import Stream, Trace
 from .accessor import ASDFAccessor, ASDFAuxiliary
 
 if TYPE_CHECKING:
-    from pypers import Space
     from pyasdf import ASDFDataSet
 
 
 # type of output data
-ASDFOutput = Optional[Union[Stream, Trace, ASDFAuxiliary, Tuple[np.array, dict]]]
+ASDFOutput = Optional[Union[Stream, Trace, ASDFAuxiliary, Tuple[np.ndarray, dict]]]
 
 
 @dataclass
@@ -69,7 +69,8 @@ class ASDFProcessor:
         if self.onerror:
             self.onerror(e)
         
-        print(format_exc(), file=stderr)
+        else:
+            print(format_exc(), file=stderr)
 
     def _open(self):
         """Open input dataset(s)."""
@@ -157,6 +158,9 @@ class ASDFProcessor:
         """Process data in current rank."""
         output = {}
 
+        # default output tag
+        output_tag = self.output_tag or self.input_tag or self.input_type
+
         for i, key in enumerate(keys):
             if i % self._size == self._rank:
                 accessors = []
@@ -168,10 +172,16 @@ class ASDFProcessor:
                 
                 # process data
                 try:
-                    output[key] = self.func(*accessors)
+                    result = self.func(*accessors)
 
-                    if isinstance(output[key], tuple):
-                        output[key] = ASDFAuxiliary(*output[key]) # type: ignore
+                    if not isinstance(result, dict):
+                        result = {output_tag: result}
+                    
+                    for key2, val in result.items():
+                        if isinstance(val, tuple):
+                            result[key2] = ASDFAuxiliary(*val)
+
+                    output[key] = result
                 
                 except Exception as e:
                     self._raise(e)
@@ -187,24 +197,23 @@ class ASDFProcessor:
         from mpi4py.MPI import COMM_WORLD as comm
         from pyasdf import ASDFDataSet
 
-        output_tag = self.output_tag or self.input_tag or self.input_type
-
         # write to output dataset
         def write():
             with ASDFDataSet(self.dst, mode='a', mpi=False, compression=None) as ds:
                 """Write to output dataset."""
-                for key, data in output.items():
-                    if isinstance(data, Stream) or isinstance(data, Trace):
-                        # write waveform data
-                        ds.add_waveforms(data, output_tag)
-                    
-                    elif isinstance(data, ASDFAuxiliary):
-                        # write auxiliary data
-                        ds.add_auxiliary_data(
-                            data = data.data,
-                            data_type = output_tag,
-                            path = key,
-                            parameters = data.parameters)
+                for key, val in output.items():
+                    for output_tag, data in val.items():
+                        if isinstance(data, Stream) or isinstance(data, Trace):
+                            # write waveform data
+                            ds.add_waveforms(data, output_tag)
+                        
+                        elif isinstance(data, ASDFAuxiliary):
+                            # write auxiliary data
+                            ds.add_auxiliary_data(
+                                data = data.data,
+                                data_type = output_tag,
+                                path = key,
+                                parameters = data.parameters)
             
         # determine the process that manages write operations
         lock_file = self.dst + '.lock'
@@ -212,9 +221,22 @@ class ASDFProcessor:
         if not exists(lock_file):
             with open(lock_file, 'a') as f:
                 f.write(str(self._rank) + ' ')
+
+        source = None
         
-        with open(lock_file, 'r') as f:
-            source = int(f.read().split(' ')[0])
+        for _ in range(3):
+            with open(lock_file, 'r') as f:
+                try:
+                    source = int(f.read().split(' ')[0])
+                
+                except Exception as e:
+                    sleep(0.1)
+                
+                else:
+                    break
+        
+        if source is None:
+            raise IOError('unable to read lock file')
         
         if source == self._rank:
             write()
